/*

pam_pop3

This module provides only the AUTH service. It checks the password with a POP3 server

Written September 2001 by Schlomo Schapiro schapiro@huji.ac.il

Version History:
10.09.2001:	Initial version


*/

#include <stdlib.h>
#include <stdio.h>
#include <syslog.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>

#define PAM_SM_AUTH
#include <security/pam_modules.h>

/* the remote port to connect to */
#define PORT 110

/* how much data to read. This should be at least one line. */
#define BUFLEN 1000

/* how much time to wait for data transfers in seconds */
#define TIMEOUT 30

/* POP3 OK and ERR strings */
#define OKSTR "+OK"
#define ERRSTR "-ERR"

/* POP3 User and Password commands including trailing space if needed !!*/
#define USERCMD "USER"
#define PASSCMD "PASS"
#define QUITCMD "QUIT"

/* POP3 end-of-line marker */
#define EOL "\r\n"

/* CHECK (and CHECK*) is set in the makefile to provide the PAM settings for standalone mode */


/* some global vars that save me typing */
int debug=0;
int port=PORT;
int timeout=TIMEOUT;
char *hostname=NULL;
char buf[BUFLEN]; /* A buffer for all kind of operations ... */


/* helper functions */

int get(int sockfd, char* desc)
/* read data from socket with timeout given in TIMEOUT and interpret it.
returns 1 for OK, 0 for ERR and -1 for some real error
but on an error it send info to syslog about the error, making use of the description.
if the external variable debug is set, it will send a debug info to syslog
*/
{
	fd_set fdset;
	struct timeval tv;
	int numbytes=0;
	int ret=0;
	FD_ZERO(&fdset);
	FD_SET(sockfd,&fdset);
	memset(buf,0,BUFLEN);
/* Wait for data from server. */
	tv.tv_sec=timeout;
	tv.tv_usec=0;
	if ((ret=select(sockfd+1,&fdset,NULL,NULL,&tv))<1)
	{
		if (ret==-1) syslog(LOG_ERR,"Error while waiting for %s from server \'%s\' port %d: %m",desc,hostname,port);
		else syslog(LOG_ERR,"Timeout after %d seconds while waiting for %s from server \'%s\' port %d",timeout,desc,hostname,port);
		close(sockfd);
		return -1;
	}
	if ((numbytes=recv(sockfd,buf,BUFLEN-1,MSG_NOSIGNAL))<=0)
	{
		if (numbytes==0)
	/* other side closed connection */
		{
			syslog(LOG_ERR,"Unexpected connection loss while reading %s from server \'%s\' port %d",desc,hostname,port);
			close(sockfd);
			return -1;
		}
		else
		{
			syslog(LOG_ERR,"Unexpected connection loss while reading %s from server \'%s\' port %d",desc,hostname,port);
			close(sockfd);
			return -1;
		}
	}
/* buf is terminated automatically from the memset */
	if ((strstr(buf,OKSTR)==NULL)&&(strstr(buf,ERRSTR)==NULL))
	{
		syslog(LOG_ERR,"Could not interpret response \'%s\' for %s from server \'%s\' port %d",buf,desc,hostname,port);
		close(sockfd);
		return -1;
	}
	if (debug) syslog(LOG_DEBUG,"Received %s: %s",desc,buf);
	return (strstr(buf,OKSTR)!=NULL) ? 1 : 0;
} /* get() */

int put(int sockfd, char* desc, char* cmd, char* arg)
/* this sends a string compesed of cmd + " " + arg + CRLF to the other side with timeout given in
TIMEOUT.
returns 0 for success and -1 for failure
at failure it also logs the failure to syslog using desc.
*/
{
	fd_set fdset;
	struct timeval tv;
	int numbytes=0;
	int ret;
	FD_ZERO(&fdset);
	FD_SET(sockfd,&fdset);
	strncpy(buf,cmd,BUFLEN-1);
	strncat(buf," ",BUFLEN-1-strlen(buf));
	strncat(buf,arg,BUFLEN-1-strlen(buf));
	strncat(buf,EOL,BUFLEN-1-strlen(buf));
	tv.tv_sec=timeout;
	tv.tv_usec=0;
	if ((ret=select(sockfd+1,NULL,&fdset,NULL,&tv)) <1)
	{
		if (ret==-1) syslog(LOG_ERR,"Error while waiting for %s write to server \'%s\' port %d: %m",desc,hostname,port);
		else syslog(LOG_ERR,"Timeout after %d seconds while waiting for %s write to server \'%s\' port %d",timeout,desc,hostname,port);
		close(sockfd);
		return -1;
	}
	if ((numbytes=send(sockfd,buf,strlen(buf),MSG_NOSIGNAL))<strlen(buf)) 
	{
/* it did not send everything, try once more and then fail */
		if (numbytes>0)
		{
			tv.tv_sec=timeout;
			tv.tv_usec=0;
		if ((ret=select(sockfd+1,NULL,&fdset,NULL,&tv)) <1)
		{
			if (ret==-1) syslog(LOG_ERR,"Error while waiting for %s write to server \'%s\' port %d: %m",desc,hostname,port);
			else syslog(LOG_ERR,"Timeout after %d seconds while waiting for %s write to server \'%s\' port %d",timeout,desc,hostname,port);
			close(sockfd);
			return -1;
		}
			/* send remaining bytes */
			numbytes+=send(sockfd,buf+numbytes,strlen(buf)-numbytes,MSG_NOSIGNAL);
		}
		if (numbytes!=strlen(buf))
		{
			syslog(LOG_ERR,"Error while writing %s to server \'%s\' port %d: %m",desc,hostname,port);
			close(sockfd);
			return -1;
		}
	}
	if (debug) syslog(LOG_DEBUG,"Sent %s: %s",desc,buf);
	return 0;
} /* put() */


/* the main PAM auth function */

PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
	int result=PAM_AUTH_ERR;
	int i,info=0;
	int sockfd,ret;
	int use_first_pass=0;
	int try_first_pass=0;
	char* val=NULL;
	struct hostent *he;
	struct sockaddr_in their_addr; //remote address info
	char *username = NULL;
	char *password = NULL;
	char *pwprompt="Password: ";

	
#ifdef DEBUG
	debug=1;
#endif
	
/* syslog stuff */
	i=LOG_CONS | LOG_PID;

#ifdef CHECK
	i|=LOG_PERROR;
#endif
	openlog("PAM-pop3",i,LOG_AUTH);
	if (debug) syslog(LOG_DEBUG,"Starting module");

/* parse options */
	for (i=0;i<argc;i++)
	{ 
		if ((val=index(argv[i],'=')) == NULL)
		/* it's a keyword, not a name=value option */
		{
			if (strcmp(argv[i],"debug")==0) debug=1;
			else if (strcmp(argv[i],"info")==0) info=1;
			else if (strcmp(argv[i],"no_warn")==0) /* don't do anything */;
			else if (strcmp(argv[i],"use_first_pass")==0) use_first_pass=1;
			else if (strcmp(argv[i],"try_first_pass")==0) try_first_pass=1;
			else if (strcmp(argv[i],"use_mapped_pass")==0) /* don't do anything */;
			else if (strcmp(argv[i],"expose_account")==0) /* don't do anything */;
			else syslog(LOG_ERR,"Unknown keyword found: %s",argv[i]);
			if (debug) syslog(LOG_DEBUG,"Got Keyword \'%s\'",argv[i]);
		}
		else
		{
		/* it's a value=keyword thing */
			/* extract the keyword into buf */
			if (strlen(argv[i])>=(BUFLEN-1))
			{
				/* we got a VERY long argument line ! */
				syslog(LOG_ALERT,"Argument too long: %s",argv[i]);
				return result;
			}
			ret=strlen(argv[i])-strlen(val);
			strncpy(buf,argv[i],ret);
			/* terminate keyword and skip over the = */
			buf[ret]=0;
			val++;
			if (debug) syslog(LOG_DEBUG,"Got Paramter \'%s\' = \'%s\'",buf,val);
			if (strcmp(buf,"hostname")==0) hostname=val;
			else if (strcmp(buf,"port")==0) port=atoi(val);
			else if (strcmp(buf,"timeout")==0) timeout=atoi(val);
			else if (strcmp(buf,"username")==0) username=val;
			else if (strcmp(buf,"password")==0) password=val;
			else if (strcmp(buf,"pwprompt")==0) pwprompt=val;
			else syslog(LOG_ERR,"Unknown keyword/value found: %s=%s",buf,val);
		}
	} /* for parse params */
	
/* get username from PAM library */
	if (username==NULL)
	{
		if (pam_get_user(pamh,(const char **)&username,NULL)!=PAM_SUCCESS)
		{
			syslog(LOG_ERR,"Could not get username from libpam !");
			return result;
		}
		if (debug) syslog(LOG_DEBUG,"Set Username to: %s",username);
	}
/* get username either from PAM library or from application */
	if (password==NULL)
	{
		if (pam_get_item(pamh,PAM_AUTHTOK,(const void **)&password)!=PAM_SUCCESS)
		{
			if (use_first_pass==1) 
			{
				if (debug) syslog(LOG_DEBUG,"Authentication failed because I did not get a password from libpam and use_first_pass is set.");
				return result;
			}
		}
	}
	if ((password==NULL) && (use_first_pass==0))
	/* did not get a password and I am allowed to ask for a new one */
	{
		struct pam_conv *conv;
		struct pam_message msg[1],*pmsg[1];
		struct pam_response *resp;
		pmsg[0]=&msg[0];
		msg[0].msg_style=PAM_PROMPT_ECHO_OFF;
		msg[0].msg=pwprompt;
		
		if (pam_get_item(pamh,PAM_CONV,(const void **)&conv)==PAM_SUCCESS)
		{
			if (ret=conv->conv(1,(const struct pam_message **)pmsg,&resp,conv->appdata_ptr)==PAM_SUCCESS)
			{
				if (strlen(resp[0].resp)>=(BUFLEN-1))
				{
					syslog(LOG_ERR,"Password given by user is too long !");
					return result;
				}
/* we got a password from the user. Store it with the application for other modules */
				if (pam_set_item(pamh,PAM_AUTHTOK,resp[0].resp)!=PAM_SUCCESS)
				{
					syslog(LOG_ERR,"Can't set password in libpam !");
					return result;
				}
				if (pam_get_item(pamh,PAM_AUTHTOK,(const void **)&password)!=PAM_SUCCESS)
				{
					if (use_first_pass==1) 
					{
						if (debug) syslog(LOG_DEBUG,"Authentication failed because I did not get a password from libpam and use_first_pass is set.");
						return result;
					}
				}
			}
			else
			{
				syslog(LOG_ERR,"Could not converse with application !");
			}
		}
		else
		{
			syslog(LOG_ERR,"Could not get password conversion function from application.");
			return result;
		}
	}
/* check wether we got all necessary infos to do the authentication */
	if ((hostname==NULL) || (username==NULL) || (password==NULL))
	{
		if (debug) syslog(LOG_ERR,"Not enough information. Need at least hostname, username and password !");
		return result;
	}

/* now we got all the info we need. Let's talk to the server */		
	if ((he=gethostbyname(hostname)) == NULL)
	{
		syslog(LOG_ALERT,"Could not translate hostname \'%s\': %s",hostname,hstrerror(h_errno));
		return result;
	}
	
	if ((sockfd = socket(AF_INET, SOCK_STREAM,0)) == -1)
	{
		syslog(LOG_ALERT,"Could not open a socket: %m");
		return result;
	}
	
/* put together the address struct */
	their_addr.sin_family = AF_INET;
	their_addr.sin_port = htons(port);
	their_addr.sin_addr = *((struct in_addr *)he->h_addr);
	memset(&(their_addr.sin_zero),0,8);
	
	if (connect(sockfd, (struct sockaddr *) &their_addr, sizeof(struct sockaddr)) == -1 )
	{
		syslog(LOG_ERR,"Could not connect to server \'%s\' port %d: %m",hostname,port);
		close(sockfd);
		return result;
	}

/* Start talking to the POP server */
	if (debug) syslog(LOG_DEBUG,"Connected to server \'%s\' port %d",hostname,port);
	switch (get(sockfd,"POP3 Greeting"))
	{
		case 1:	break;
		case 0:	syslog(LOG_ERR,"Got bad Greeting from server \'%s\' port %d: %m",hostname,port);
		default:
			return result;
	} 
	
/* here we got a POP3 greeting. Send username. */
	if (put(sockfd,"Username",USERCMD,username)!=0) return result;
/* wait and check for OK on the username */
	switch (get(sockfd,"Username Response"))
	{
		case 1:	break;
		case 0:	syslog(LOG_ERR,"Got bad Username Response from server \'%s\' port %d: %m",hostname,port);
		default:
			return result;
	}
/* here we got the OK on the username, send password and evaluate result */
	if (put(sockfd,"Password",PASSCMD,password)!=0)	return result;
/* wait for password OK */
	if ((i=get(sockfd,"Password Response"))==-1) return result;
/* close connection gracefully */
	put(sockfd,"QUITting",QUITCMD,"");
	close(sockfd);
	if (debug || info) 
	{
		if (i==1) syslog(LOG_DEBUG,"Authentication Succeeded for %s at server %s port %d",username,hostname,port);
		else syslog(LOG_DEBUG,"Authentication Failed for %s at server %s port %d",username,hostname,port);
	}
/* return PAM_SUCCESS if we got OK */
	return (i==1) ? PAM_SUCCESS : result;
}

PAM_EXTERN int pam_sm_setcred(pam_handle_t *pamh,int flags, int argc, const char **argv)
/* this is recommended to do the same as the auth function, don't ask me why ! */
/* but in pam_rootok they just return PAM_SUCCESS and since the auth line in pam.d runs
   both functions, we dot the latter */
{
	return PAM_SUCCESS;
/*	return pam_sm_authenticate(pamh,flags,argc,argv);*/
}

int main(int argc, const char **argv)
{
/* call pam function. Remove 0th parameter since it is the program name while if called from libpam
the 0th parameter is significant */
	if (pam_sm_authenticate(NULL,0,argc-1,argv+1) == PAM_SUCCESS) 
		printf("Authentication succeeded\n");
	else
		printf("Authentication failed\n");
	return 0;
}
